local _DEBUG = false
local _UPDATEDISPLAY_RATE = 0.2
local _SCANTALENT_RATE = 1

ShadowGreenLight = LibStub("AceAddon-3.0"):NewAddon("ShadowGreenLight", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local TalentQuery = LibStub:GetLibrary("LibTalentQuery-1.0")

-- Option table
local options = { 
    name = "ShadowGreenLight",
    handler = ShadowGreenLight,
    type = "group",
    args = {
        reset = {
            type = "execute",
            name = "Reset position",
            desc = "Reset display position",
            func = "ResetDisplayPosition",
        },
    },
}

function ShadowGreenLight:OnInitialize()
  	-- Initialization of attributes
  	ShadowGreenLight.maxUpgrade = 0
	ShadowGreenLight.currentUpgrade = 0
	ShadowGreenLight.upgradeList = {}
	ShadowGreenLight.currentUpgradeList = {}
	ShadowGreenLight.raidUpgradeAvailable = {}
	ShadowGreenLight.scanStatus = {}
	
	-- Initialization of spell database
	ShadowGreenLight.upgradeDb = {
  		["ShadowWeave"] = {
  			name = GetSpellInfo(15257),
			maxCount = 5,
			upgradePerStack = 2,
			type = "Shadow Weaving",
			detectionType = "playerBuffs",
		},
		["Improved Scorch"] = {
  			name = GetSpellInfo(11095),
  			talentName = GetSpellInfo(11095),
  			talentTab = 2,
  			talentIndex = 11,
			maxCount = 5,
			upgradePerStack = 2,
			type = "Spell Critical Strike Chance Debuff",
			detectionType = "targetDebuffs",
		},
		["Winter's Chill"] = {
  			name = GetSpellInfo(11180),
  			talentName = GetSpellInfo(11180),
  			talentTab = 3,
  			talentIndex = 18,
			maxCount = 5,
			upgradePerStack = 2,
			type = "Spell Critical Strike Chance Debuff",
			detectionType = "targetDebuffs",
		},
		["Totem of Wrath"] = {
  			name = GetSpellInfo(57722),
  			talentName = GetSpellInfo(57722),
  			talentTab = 1,
  			talentIndex = 21,
			maxCount = 1,
			upgradePerStack = 3,
			type = "Critical Strike Chance Taken Debuff",
			detectionType = "targetDebuffs",
		},
		["Heart of the crusader"] = {
  			name = GetSpellInfo(20337),
  			talentName = GetSpellInfo(20337),
  			talentTab = 3,
  			talentIndex = 4,
			maxCount = 1,
			upgradePerTalentRank = {
				[1] = 1,
				[2] = 2,
				[3] = 3,
			},
			type = "Critical Strike Chance Taken Debuff",
			detectionType = "targetDebuffs",
		},
		["Moonkin Aura"] = {
  			name = GetSpellInfo(24907),
  			talentName = GetSpellInfo(24858),
  			talentTab = 1,
  			talentIndex = 18,
			maxCount = 1,
			upgradePerStack = 5,
			type = "Spell Critical Strike Chance Buff",
			detectionType = "playerBuffs",
		},
		["Elemental Oath"] = {
  			name = GetSpellInfo(53414),
  			talentName = GetSpellInfo(53414),
  			talentTab = 1,
  			talentIndex = 18,
			maxCount = 1,
			upgradePerTalentRank = {
				[1] = 3,
				[2] = 5,
			},
			type = "Spell Critical Strike Chance Buff",
			detectionType = "playerBuffs",
		},
  	}
  	
  	-- Initialization of session db
  	self.db = LibStub("AceDB-3.0"):New("ShadowGreenLightDB", {
  		profile = {
  			posX = 0,
  			posY = 0,
  			relativePoint = "CENTER",
  			point = "CENTER",
		}
	}, "Default")
	
	-- Registering chat commands
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ShadowGreenLight", options, {"ShadowGreenLight", "gl"})

end

function ShadowGreenLight:OnEnable()
	-- Registering event for group / raid joining
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "RosterUpdate")
	-- Registering events to enter/exit combat
    self:RegisterEvent("PLAYER_REGEN_DISABLED", "EnteringCombat")
    self:RegisterEvent("PLAYER_REGEN_ENABLED", "ExitingCombat")
    -- Callback for talent query
    TalentQuery.RegisterCallback(self, "TalentQuery_Ready")
    -- Create display
    self:CreateDisplay()
end

function ShadowGreenLight:OnDisable()
    -- Callback for talent query
    TalentQuery.UnregisterCallback(self, "TalentQuery_Ready")
end

function ShadowGreenLight:CreateDisplay()
	self.frame = CreateFrame("Frame", "ShadowGreenLight_Frame", UIParent)
	self.frame:SetHeight(30)
	self.frame:SetWidth(30)
	self.frame:SetPoint(self.db.profile.point,UIParent,self.db.profile.relativePoint,self.db.profile.posX,self.db.profile.posY)
	self.frame:EnableMouse()
	self.frame:SetMovable(true)
	self.frame:SetResizable(true)
	self.frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "",
		tile = false,
		tileSize = 0,
		edgeSize = 0,
		insets = {left = 0, right = 0, top = 0, bottom = 0},
	})
	self.frame:SetBackdropColor(0,0,0,1)
	
	self.frame:SetScript("OnMouseDown", function() 
		if ( ( ( not this.isLocked ) or ( this.isLocked == 0 ) ) and ( arg1 == "RightButton" ) ) then
			this:StartMoving();
			this.isMoving = true;
		end
	end)
	self.frame:SetScript("OnMouseUp", function() 
		if ( this.isMoving ) then
			local point,relativeTo,relativePoint,xOfs,yOfs = this:GetPoint(1)
			self.db.profile.point = point
			self.db.profile.relativePoint = relativePoint
			self.db.profile.posX = xOfs
			self.db.profile.posY = yOfs
			this:StopMovingOrSizing();
			this.isMoving = false;
		end
	end)
end

function ShadowGreenLight:ResetDisplayPosition()
	self:Debug("Display position reset")
	self.frame:ClearAllPoints()
	self.frame:SetPoint("CENTER",UIParent,"CENTER", 0, 0)
end

function ShadowGreenLight:Debug(msg, force)
	if _DEBUG or force then
		self:Print(msg)
	end
end

function ShadowGreenLight:RosterUpdate()
	-- Schedule timer to scan talents if in raid or group
	if (GetNumRaidMembers() > 0 or GetNumPartyMembers() > 0) then
		if not self.talentScanTimer then
			self.talentScanTimer = self:ScheduleRepeatingTimer("ScanRaidTalents", _SCANTALENT_RATE)
			self:Debug("Scheduling timer for talent scan")
		end
	-- Unschedule timer to scan talents if not in a raid or group
	elseif self.talentScanTimer then
		self:CancelTimer(self.talentScanTimer)
		self.talentScanTimer = nil
		self:Debug("Unscheduling timer for talent scan")
	end
end

function ShadowGreenLight:EnteringCombat()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateDisplay")
	self.updateDisplayTimer = self:ScheduleRepeatingTimer("UpdateDisplay", _UPDATEDISPLAY_RATE)
	
	self:PopulateUpgradeList()
	self:ComputeMaxUpgrade()
end

function ShadowGreenLight:ExitingCombat()
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:CancelTimer(self.updateDisplayTimer)
	self.updateDisplayTimer = nil
	
	self.frame:SetBackdropColor(0,0,0,1)
end

---------------------------------------------------
-- Talent scanning

function ShadowGreenLight:ScanRaidTalents()
	if GetNumRaidMembers()>0 then
		for i = 1,40 do
			self:ScanUnitIfNotReceived("raid"..i)
		end
	else
		if GetNumPartyMembers() > 0 then
			for i = 1,4 do
				self:ScanUnitIfNotReceived("party"..i)
			end
		end
	end
end

function ShadowGreenLight:ScanUnitIfNotReceived(unit)
	if UnitExists(unit) then
		if not self.scanStatus[UnitName(unit)] or self.scanStatus[UnitName(unit)] ~= "RECEIVED" then
			self.scanStatus[UnitName(unit)] = "SENT"
			if UnitName(unit)==UnitName("player") then
				self:TalentQuery_Ready(_, UnitName(unit))
			else
				TalentQuery:Query(unit)
			end
			return true
		end
	end
	return false
end

-- Handle LibTalentQuery-1.0 TalentQuery_Ready event
function ShadowGreenLight:TalentQuery_Ready(_, name)
	self:Debug("Talent info received for " .. name)
	local isNotPlayer = (UnitName("player")~=name)
	-- Scan spells in upgrade database (except shadowweave)
	for upgradeName, upgrade in pairs(self.upgradeDb) do
		-- skip ShadowWeave
		if upgrade.talentName then
			-- Get talent related to upgrade 
			local talentName, _, _, _, talentRank = GetTalentInfo(upgrade.talentTab, upgrade.talentIndex, isNotPlayer)
			if talentName and talentName == upgrade.talentName then
				if talentRank > 0 then
					if not self.raidUpgradeAvailable[name] then
						self.raidUpgradeAvailable[name] = {}
					end
					if not self.raidUpgradeAvailable[name][upgradeName] then
						self.raidUpgradeAvailable[name][upgradeName] = {
							name = upgrade.name,
				  			maxCount = upgrade.maxCount,
							upgradePerStack = upgrade.upgradePerStack,
							type = upgrade.type,
							detectionType = upgrade.detectionType,
						}
						-- upgrade per stack can be dependant on talent point investment
						if upgrade.upgradePerTalentRank then
							self.raidUpgradeAvailable[name][upgradeName].upgradePerStack = upgrade.upgradePerTalentRank[talentRank]
						end
					end
					self:Debug(name .. " > " .. upgradeName .. " > " .. self.raidUpgradeAvailable[name][upgradeName].upgradePerStack)
				end 
			end
		end
	end
	self.scanStatus[name] = "RECEIVED"
end

---------------------------------------------------
-- Update Display
function ShadowGreenLight:UpdateDisplay()
	-- Reset current upgrade list
	for upgradeType in pairs(self.currentUpgradeList) do
		self.currentUpgradeList[upgradeType] = nil
	end
	
	-- Scan player buffs
	local index = 1
	local auraName, _, _, auraCount = UnitBuff("player", index)
	while auraName do
		for _, upgradeBuff in pairs(self.upgradeList["playerBuffs"]) do
			if upgradeBuff.name == auraName then
				if not self.currentUpgradeList[upgradeBuff.type] then
					self.currentUpgradeList[upgradeBuff.type] = 0
				end
				local upgradeAmount = upgradeBuff.upgradePerStack * auraCount
				-- Check if need updating
				if upgradeAmount > self.currentUpgradeList[upgradeBuff.type] then
					self.currentUpgradeList[upgradeBuff.type] = upgradeAmount
				end
			end
		end
		index = index + 1
		auraName, _, _, auraCount = UnitBuff("player", index)
	end
	
	-- Scan target debuffs
	index = 1
	auraName, auraRank, _, auraCount = UnitDebuff("target", index)
	while auraName do
		for _, upgradeBuff in pairs(self.upgradeList["targetDebuffs"]) do
			if upgradeBuff.name == auraName then
				if not self.currentUpgradeList[upgradeBuff.type] then
					self.currentUpgradeList[upgradeBuff.type] = 0
				end
				local upgradeAmount = upgradeBuff.upgradePerStack * auraCount
				-- Check if need updating
				if upgradeAmount > self.currentUpgradeList[upgradeBuff.type] then
					self.currentUpgradeList[upgradeBuff.type] = upgradeAmount
				end
			end
		end
		index = index + 1
		auraName, _, _, auraCount = UnitDebuff("target", index)
	end
	
	local currentUpgrade = 0
	for _, upgradeAmount in pairs(self.currentUpgradeList) do
		currentUpgrade = currentUpgrade + upgradeAmount
	end
	
	self.frame:SetBackdropColor(
		1-currentUpgrade / self.maxUpgrade,
		currentUpgrade / self.maxUpgrade,
		0,
		1
	)
end

---------------------------------------------------
-- Populate Upgrade List

function ShadowGreenLight:PopulateUpgradeList()
	-- reinitialize upgrade list
	if not self.upgradeList["playerBuffs"] then
		self.upgradeList["playerBuffs"] = {}
	end
	if not self.upgradeList["targetDebuffs"] then
		self.upgradeList["targetDebuffs"] = {}
	end
	for detectionType, _ in pairs(self.upgradeList) do
		for upgradeName, _ in pairs(self.upgradeList[detectionType]) do
			self.upgradeList[detectionType][upgradeName] = nil
		end
	end
	
	-- Always assume user is a shadowpriest with maxed shadow weaving (duh ?)
	self.upgradeList["playerBuffs"]["ShadowWeave"] = self.upgradeDb["ShadowWeave"]
	
	-- taking list of raid member and comparing it to available upgrade list
	if GetNumRaidMembers()>0 then
		for i = 1,40 do
			self:PopulateUpgradeListForUnit("raid"..i)
		end
	else
		if GetNumPartyMembers() > 0 then
			for i = 1,4 do
				self:PopulateUpgradeListForUnit("party"..i)
			end
		end
	end
end

function ShadowGreenLight:PopulateUpgradeListForUnit(unit)
	local name = UnitName(unit)
	if self.raidUpgradeAvailable[name] then
		for upgradeName, upgrade in pairs(self.raidUpgradeAvailable[name]) do
			-- either buff was not available before
			if not self.upgradeList[upgrade.detectionType][upgradeName] then
				self.upgradeList[upgrade.detectionType][upgradeName] = upgrade
			-- or it was, in that case check if new upgrade is better
			else
				if self.upgradeList[upgrade.detectionType][upgradeName].upgradePerStack < upgrade.upgradePerStack then
					self.upgradeList[upgrade.detectionType][upgradeName] = upgrade
				end
			end
		end
	end
end

function ShadowGreenLight:ComputeMaxUpgrade()
	local maxUpgradeByUpgradeType = {}
	self.maxUpgrade = 0
	for detectionType, upgradeArray in pairs(self.upgradeList) do
		for _, upgradeBuff in pairs(upgradeArray) do
			if not maxUpgradeByUpgradeType[upgradeBuff.type] then
				maxUpgradeByUpgradeType[upgradeBuff.type] = 0
			end
			local maxUpgradeAmount = upgradeBuff.upgradePerStack * upgradeBuff.maxCount
			if maxUpgradeAmount > maxUpgradeByUpgradeType[upgradeBuff.type] then
				maxUpgradeByUpgradeType[upgradeBuff.type] = maxUpgradeAmount
			end
		end
	end
	
	self:Debug("***************")
	for upgradeType, maxUpgradeAmount in pairs(maxUpgradeByUpgradeType) do
		self.maxUpgrade = self.maxUpgrade + maxUpgradeAmount
		self:Debug(upgradeType .. " > " .. maxUpgradeAmount)
	end
end
